"""Creates the Windfarm class/model."""
from __future__ import annotations

from math import fsum
from itertools import chain, combinations

import numpy as np
import pandas as pd  # type: ignore
import networkx as nx  # type: ignore
from geopy import distance  # type: ignore

from wombat.core import RepairManager, WombatEnvironment
from wombat.core.library import load_yaml
from wombat.windfarm.system import Cable, System


try:  # pylint: disable=duplicate-code
    from functools import cache  # type: ignore
except ImportError:  # pylint: disable=duplicate-code
    from functools import lru_cache  # pylint: disable=duplicate-code

    cache = lru_cache(None)  # pylint: disable=duplicate-code


class Windfarm:
    """The primary class for operating on objects within a windfarm. The substations,
    cables, and turbines are created as a network object to be more appropriately accessed
    and controlled.
    """

    def __init__(
        self,
        env: WombatEnvironment,
        windfarm_layout: str,
        repair_manager: RepairManager,
    ) -> None:

        self.env = env
        self.repair_manager = repair_manager

        # Set up the layout and instantiate all windfarm objects
        self._create_graph_layout(windfarm_layout)
        self._create_turbines_and_substations()
        self._create_cables()
        self.capacity = sum(self.system(turb).capacity for turb in self.turbine_id)
        self._create_substation_turbine_map()
        self.calculate_distance_matrix()

        # Create the logging items
        self.system_list = list(self.graph.nodes)
        self._log_columns = [
            "datetime",
            "name",
            "level",
            "env_datetime",
            "env_time",
        ] + self.system_list

        # Register the windfarm and start the logger
        self.repair_manager._register_windfarm(self)
        self.env.process(self._log_operations())

    def _create_graph_layout(self, windfarm_layout: str) -> None:
        """Creates a network layout of the windfarm start from the substation(s) to
        be able to capture downstream turbines that can be cut off in the event of a cable failure.

        Parameters
        ----------
        windfarm_layout : str
            Filename to use for reading in the windfarm layout; must be a csv file.
        """
        layout_path = str(self.env.data_dir / "windfarm" / windfarm_layout)
        layout = (
            pd.read_csv(layout_path)
            .sort_values(by=["string", "order"])
            .reset_index(drop=True)
        )
        layout.subassembly = layout.subassembly.fillna("")
        layout.upstream_cable = layout.upstream_cable.fillna("")

        windfarm = nx.DiGraph()
        windfarm.add_nodes_from(layout.id.values)

        # Assign the data attributes to the graph nodes
        for col in ("name", "latitude", "longitude", "subassembly"):
            d = {i: n for i, n in layout[["id", col]].values}
            nx.set_node_attributes(windfarm, d, name=col)

        # Determine which nodes are substations and which are turbines
        substation_filter = layout.id == layout.substation_id
        self.substation_id = layout[substation_filter].id.unique()
        self.turbine_id = layout[~substation_filter].id.unique()
        _type = {True: "substation", False: "turbine"}
        d = {i: _type[val] for i, val in zip(layout.id, substation_filter.values)}
        nx.set_node_attributes(windfarm, d, name="type")

        substations = layout[substation_filter].copy()
        turbines = layout[~substation_filter].copy()
        substation_sections = [
            turbines[turbines.substation_id == substation]
            for substation in substations.id
        ]
        for section in substation_sections:
            for _, row in section.iterrows():
                if row.order == 0:
                    start: str = row.substation_id
                else:
                    start = current  # noqa: F821
                current: str = row.id
                windfarm.add_edge(
                    start, current, length=row.distance, cable=row.upstream_cable
                )

        self.graph = windfarm

    def _create_turbines_and_substations(self) -> None:
        for system_id, data in self.graph.nodes(data=True):
            if data["subassembly"] == "":
                raise ValueError(
                    "A 'subassembly' file must be specified for all nodes in the windfarm layout!"
                )

            subassembly_dict = load_yaml(
                self.env.data_dir / "windfarm", data["subassembly"]
            )
            self.graph.nodes[system_id]["system"] = System(
                self.env,
                self.repair_manager,
                system_id,
                data["name"],
                subassembly_dict,
                data["type"],
            )

    def _create_cables(self) -> None:
        for start_node, end_node, data in self.graph.edges(data=True):
            # Check that the cable data is provided
            if data["cable"] == "":
                raise ValueError(
                    "A 'cable' file must be specified for all nodes in the windfarm layout!"
                )
            cable_dict = load_yaml(self.env.data_dir / "windfarm", data["cable"])

            start_coordinates = (
                self.graph.nodes[start_node]["latitude"],
                self.graph.nodes[start_node]["longitude"],
            )
            end_coordinates = (
                self.graph.nodes[end_node]["latitude"],
                self.graph.nodes[end_node]["longitude"],
            )

            # If the real distance/cable length is not input, then the geodesic distance
            # is calculated
            if data["length"] == 0:
                data["length"] = distance.geodesic(
                    start_coordinates, end_coordinates, ellipsoid="WGS-84"
                ).km

            data["cable"] = Cable(
                self,
                self.env,
                start_node,
                end_node,
                cable_dict,
            )

            # Calaculate the geometric center point
            end_points = np.array((start_coordinates, end_coordinates))
            data["latitude"], data["longitude"] = end_points.mean(axis=0)

    def calculate_distance_matrix(self) -> None:
        """Calculates hte geodesic distance, in km, between all of the windfarm's nodes, e.g.,
        substations and turbines, and cables.
        """
        ids = list(self.graph.nodes())
        ids.extend([data["cable"].id for *_, data in self.graph.edges(data=True)])
        coords = [
            (data["latitude"], data["longitude"])
            for *_, data in (*self.graph.nodes(data=True), *self.graph.edges(data=True))
        ]

        dist = [distance.geodesic(c1, c2).km for c1, c2 in combinations(coords, 2)]
        dist_arr = np.ones((len(ids), len(ids)))
        triangle_ix = np.triu_indices_from(dist_arr, 1)
        dist_arr[triangle_ix] = dist_arr.T[triangle_ix] = dist

        # Set the self distance to infinity, so that only one crew can be dropped off
        # at a single point
        np.fill_diagonal(dist_arr, np.inf)

        self.distance_matrix = pd.DataFrame(dist_arr, index=ids, columns=ids)

    def _create_substation_turbine_map(self) -> None:
        """Creates ``substation_turbine_map``, a dictionary, that maps substation(s) to
        the dependent turbines in the windfarm, and the weighting of each turbine in the
        windfarm.
        """
        # Get all turbines dependent on each substation
        s_t_map = {
            s_id: list(
                chain.from_iterable(nx.dfs_successors(self.graph, source=s_id).values())
            )
            for s_id in self.substation_id
        }

        # Reorient the mapping to have the turbine list and the capacity-based weighting
        # of each turbine
        s_t_map = {
            s_id: dict(  # type: ignore
                turbines=np.array(s_t_map[s_id]),
                weights=np.array([self.system(t).capacity for t in s_t_map[s_id]])
                / self.capacity,
            )
            for s_id in s_t_map
        }
        self.substation_turbine_map = s_t_map

    def _log_operations(self):
        """Logs the operational data for a simulation."""
        self.env._operations_logger.info(" :: ".join(self._log_columns))

        message = [self.env.simulation_time, self.env.now]
        message.extend(
            [self.system(system).operating_level for system in self.system_list]
        )
        message = " :: ".join((f"{m}" for m in message))  # type: ignore
        self.env._operations_logger.info(message)

        HOURS = 1
        while True:
            yield self.env.timeout(HOURS)
            message = [f"{self.env.simulation_time}", f"{self.env.now}"] + [
                f"{self.system(system).operating_level}" for system in self.system_list
            ]
            message = " :: ".join(message)  # type: ignore
            self.env._operations_logger.info(message)

    @cache
    def system(self, system_id: str) -> System:
        """Convenience function to returns the desired `System` object for a turbine or
        substation in the windfarm.

        Parameters
        ----------
        system_id : str
            The system's unique identifier, ``wombat.windfarm.System.id``.

        Returns
        -------
        System
            The ``System`` object.
        """
        return self.graph.nodes[system_id]["system"]

    @cache
    def cable(self, cable_id: tuple[str, str] | str) -> Cable:
        """Convenience function to returns the desired `Cable` object for a cable in the
        windfarm.

        Parameters
        ----------
        cable_id : tuple[str, str] | str
            The cable's unique identifier, of the form: (``wombat.windfarm.System.id``,
            ``wombat.windfarm.System.id``), for the (downstream node id, upstream node id),
            or the ``Cable.id``.

        Returns
        -------
        Cable
            The ``Cable`` object.
        """
        if isinstance(cable_id, str):
            edge_id = tuple(cable_id.split("::")[1:])
        else:
            edge_id = cable_id
        try:
            return self.graph.edges[edge_id]["cable"]
        except KeyError:
            raise KeyError(f"Edge {edge_id} is invalid.")

    @property
    def current_availability(self) -> float:
        """Calculates the product of all system ``operating_level`` variables across the
        windfarm using the following forumation

        .. math::
            \sum{
                OperatingLevel_{substation_{i}} *
                \sum{OperatingLevel_{turbine_{j}} * Weight_{turbine_{j}}}
            }

        where the :math:``{OperatingLevel}`` is the product of the operating level
        of each subassembly on a given system (substation or turbine), and the
        :math:``{Weight}`` is the proportion of one turbine's capacity relative to
        the whole windfarm.
        """  # noqa: W605
        operating_levels = {
            s_id: [
                self.system(t).operating_level
                for t in self.substation_turbine_map[s_id]["turbines"]  # type: ignore
            ]
            for s_id in self.substation_turbine_map
        }
        availability = fsum(
            [
                self.system(s_id).operating_level
                * fsum(
                    operating_levels[s_id]
                    * self.substation_turbine_map[s_id]["weights"]  # type: ignore
                )
                for s_id in self.substation_turbine_map
            ]
        )
        return availability

    @property
    def current_availability_wo_servicing(self) -> float:
        """Calculates the product of all system ``operating_level`` variables across the
        windfarm using the following forumation, ignoring 0 operating level due to ongoing
        servicing.

        .. math::
            \sum{
                OperatingLevel_{substation_{i}} *
                \sum{OperatingLevel_{turbine_{j}} * Weight_{turbine_{j}}}
            }

        where the :math:``{OperatingLevel}`` is the product of the operating level
        of each subassembly on a given system (substation or turbine), and the
        :math:``{Weight}`` is the proportion of one turbine's capacity relative to
        the whole windfarm.
        """  # noqa: W605
        operating_levels = {
            s_id: [
                self.system(t).operating_level_wo_servicing
                for t in self.substation_turbine_map[s_id]["turbines"]  # type: ignore
            ]
            for s_id in self.substation_turbine_map
        }
        availability = fsum(
            [
                self.system(s_id).operating_level_wo_servicing
                * fsum(
                    operating_levels[s_id]
                    * self.substation_turbine_map[s_id]["weights"]  # type: ignore
                )
                for s_id in self.substation_turbine_map
            ]
        )
        return availability



"""Creates the ServiceEquipment class for simulating repairs."""
# TODO: NEED A SPECIFIC STARTUP METHOD
from __future__ import annotations

from math import ceil
from typing import Generator  # type: ignore
from datetime import timedelta

import numpy as np  # type: ignore
from simpy.events import Process, Timeout  # type: ignore
from pandas.core.indexes.datetimes import DatetimeIndex

from wombat.core import (
    Failure,
    Maintenance,
    RepairManager,
    RepairRequest,
    WombatEnvironment,
    ServiceEquipmentData,
)
from wombat.windfarm import Windfarm
from wombat.utilities import hours_until_future_hour
from wombat.core.library import load_yaml
from wombat.windfarm.system import System
from wombat.core.data_classes import ScheduledServiceEquipmentData
from wombat.windfarm.system.cable import Cable
from wombat.windfarm.system.subassembly import Subassembly


try:
    from functools import cache  # type: ignore
except ImportError:
    from functools import lru_cache

    cache = lru_cache(None)


HOURS_IN_DAY = 24


def consecutive_groups(data: np.ndarray, step_size: int = 1) -> list[np.ndarray]:
    """Generates the subgroups of an array where the difference between two sequential
    elements is equal to the ``step_size``. The intent is to find the length of delays
    in a weather forecast

    Parameters
    ----------
    data : np.ndarray
        An array of integers.
    step_size : int, optional
        The step size to be considered a consecutive number, by default 1.

    Returns
    -------
    list[np.ndarray]
        A list of arrays of the consecutive elements of ``data``.
    """
    # Consecutive groups in delay
    groups = np.split(data, np.where(np.diff(data) != step_size)[0] + 1)
    return groups


class ServiceEquipment:
    """Provides a servicing equipment object that can handle various maintenance and
    repair tasks.

    Parameters
    ----------
    env : WombatEnvironment
        The simulation environment.
    windfarm : Windfarm
        The ``Windfarm`` object.
    repair_manager : RepairManager
        The ``RepairManager`` object.
    equipment_data_file : str
        The equipment settings file name with extension.
    """

    def __init__(
        self,
        env: WombatEnvironment,
        windfarm: Windfarm,
        repair_manager: RepairManager,
        equipment_data_file: str,
    ):
        """Initializes the ``ServiceEquipment`` class.

        Parameters
        ----------
        env : WombatEnvironment
            The simulation environment.
        windfarm : Windfarm
            The ``Windfarm`` object.
        repair_manager : RepairManager
            The ``RepairManager`` object.
        equipment_data_file : str
            The equipment settings file name with extension.
        """
        self.env = env
        self.windfarm = windfarm
        self.manager = repair_manager
        self.onsite = False
        self.enroute = False
        self.at_port = False
        self.at_system = False
        self.transferring_crew = False
        self.current_system = None  # type: str | None

        data = load_yaml(env.data_dir / "repair" / "transport", equipment_data_file)
        try:
            if data["start_year"] < self.env.start_year:
                data["start_year"] = self.env.start_year
            if data["end_year"] > self.env.end_year:
                data["end_year"] = self.env.end_year
        except KeyError:
            # Ignores for unscheduled maintenace equipment that would not have this input
            pass

        # NOTE: mypy is not caught up with attrs yet :(
        self.settings = ServiceEquipmentData(data).determine_type()  # type: ignore
        self._check_working_hours()

        # Register servicing equipment with the repair manager if it is using an
        # unscheduled maintenance scenario, so it can be dispatched as needed
        if self.settings.strategy != "scheduled":
            self.manager._register_equipment(self)

        # Only run the equipment if it is on a scheduled basis, otherwise wait
        # for it to be dispatched
        if self.settings.strategy == "scheduled":
            self.env.process(self.run_scheduled())

    def _check_working_hours(self) -> None:
        """Checks the working hours of the equipment and overrides a default (-1) to
        the ``env`` settings, otherwise hours remain the same.
        """
        start_is_invalid = self.settings.workday_start == -1
        end_is_invalid = self.settings.workday_end == -1

        start = (
            self.env.workday_start if start_is_invalid else self.settings.workday_start
        )
        end = self.env.workday_end if end_is_invalid else self.settings.workday_end

        self.settings._set_environment_shift(start, end)

    def calculate_salary_cost(self, duration: int | float) -> float:
        """The total salaried labor cost implications for a given action.

        Parameters
        ----------
        duration : int | float
            Number of hours to charge for the salaried labor.

        Returns
        -------
        float
            The total cost of salaried workers for a given action.
        """
        return (
            (duration / HOURS_IN_DAY)
            * self.settings.crew.day_rate
            * self.settings.crew.n_day_rate
        )

    def calculate_hourly_cost(self, duration: int | float) -> float:
        """The total hourly labor cost implications for a given action.

        Parameters
        ----------
        duration : int | float
            Number of hours to charge for the hourly labor.

        Returns
        -------
        float
            The total cost of hourly workers for a given action.
        """
        return (
            duration * self.settings.crew.hourly_rate * self.settings.crew.n_hourly_rate
        )

    def calculate_equipment_cost(self, duration: int | float) -> float:
        """The total equipment cost implications for a given action.

        Parameters
        ----------
        duration : int | float
            Number of hours to charge for the equipment.

        Returns
        -------
        float
            The total cost of equipment for a given action.
        """
        return (duration / HOURS_IN_DAY) * self.settings.equipment_rate

    def register_repair_with_subassembly(
        self,
        subassembly: Subassembly | Cable,
        repair: RepairRequest,
        starting_operating_level: float,
    ) -> None:
        """Goes into the repaired subassembly, component, or cable and returns its
        ``operating_level`` back to good as new for the specific repair. For fatal cable
        failures, all upstream turbines are turned back on unless there is another fatal
        cable failure preventing any more from operating.

        Parameters
        ----------
        subassembly : Subassembly | Cable
            The subassembly or cable that was repaired.
        repair : RepairRequest
            The request for repair that was submitted.
        starting_operating_level : float
            The operating level before a repair was started.
        """
        # If this is a cable repair and returning it back to operating, then reset all
        # upstream turbines to be operating as long as there isn't another fatal cable
        # failure preventing it from operating
        if repair.cable and repair.details.operation_reduction == 1:
            start: str = ""
            for i, turbine_id in enumerate(repair.upstream_turbines):
                if i > 0:
                    cable = self.windfarm.cable((start, turbine_id))
                    if cable.operating_level == 0:
                        break
                turbine = self.windfarm.system(turbine_id)
                turbine.cable_failure = False
                start = turbine_id

        # Put the subassembly/component back to good as new condition
        if repair.details.operation_reduction == 1:
            subassembly.operating_level = 1
            _ = self.manager.purge_subassembly_requests(
                repair.system_id, repair.subassembly_id
            )
        elif repair.details.operation_reduction == 0:
            subassembly.operating_level = starting_operating_level
        else:
            subassembly.operating_level /= 1 - repair.details.operation_reduction

    def wait_until_next_shift(self, **kwargs) -> Generator[Timeout, None, None]:
        """Delays the crew until the start of the next shift.

        Yields
        -------
        Generator[Timeout, None, None]
            Delay until the start of the next shift.
        """
        kwargs["additional"] = kwargs.get(
            "additional", "work shift has ended; waiting for next shift to start"
        )
        kwargs["action"] = kwargs.get("action", "delay")
        delay = self.env.hours_to_next_shift(workday_start=self.settings.workday_start)
        salary_cost = self.calculate_salary_cost(delay)
        hourly_cost = self.calculate_hourly_cost(0)
        equpipment_cost = self.calculate_equipment_cost(delay)
        self.env.log_action(
            duration=delay,
            salary_labor_cost=salary_cost,
            hourly_labor_cost=hourly_cost,
            equipment_cost=equpipment_cost,
            **kwargs,
        )
        yield self.env.timeout(delay)

    def wait_until_next_operational_period(
        self, *, less_mobilization_hours: int = 0
    ) -> Generator[Timeout, None, None]:
        """Delays the crew and equipment until the start of the next operational
        period.

        TODO: Need a custom error if weather doesn't align with the equipment dates.

        Parameters
        ----------
        less_mobilization_hours : int
            The number of hours required for mobilization that will be subtracted from
            the waiting period to account for mobilization, by default 0.

        Yields
        -------
        Generator[Timeout, None, None]
            A Timeout event for the number of hours between when the function is called
            and when the next operational period starts.
        """
        current = self.env.simulation_time.date()
        ix_match = np.where(current < self.settings.operating_dates)[0]  # type: ignore
        if ix_match.size > 0:
            still_in_operation = True
            next_operating_date = self.settings.operating_dates[ix_match[0]]  # type: ignore
            hours_to_next_shift = (
                self.env.date_ix(next_operating_date)
                - self.env.now
                + self.settings.workday_start
                - less_mobilization_hours
            )
        else:
            still_in_operation = False
            hours_to_next_shift = self.env.max_run_time - self.env.now

        if still_in_operation:
            # TODO: This message needs to account for unscheduled equipment as well, but
            # the post processor filtering needs to be updated as well
            additional = "will return next year"
        else:
            additional = "no more return visits will be made"

        # Ensures that the statuses are correct
        self.at_port = False
        self.enroute = False
        self.onsite = False

        self.env.log_action(
            agent=self.settings.name,
            action="delay",
            reason="work is complete",
            additional=additional,
            duration=hours_to_next_shift,
        )

        yield self.env.timeout(hours_to_next_shift)

    def mobilize_scheduled(self) -> Generator[Timeout, None, None]:
        """Mobilizes the ServiceEquipment object by waiting for the next operational
        period, less the mobilization time, then logging the mobiliztion cost.

        NOTE: weather delays are not accounted for in this process.

        Yields
        -------
        Generator[Timeout, None, None]
            A Timeout event for the number of hours between when the function is called
            and when the next operational period starts.
        """
        mobilization_hours = self.settings.mobilization_days * HOURS_IN_DAY
        yield self.env.process(
            self.wait_until_next_operational_period(
                less_mobilization_hours=mobilization_hours
            )
        )
        self.env.log_action(
            agent=self.settings.name,
            action="mobilization",
            reason=f"{self.settings.name} is being mobilized",
            additional="mobilization",
        )

        self.enroute = True
        yield self.env.timeout(mobilization_hours)
        self.onsite = True
        self.enroute = False

        self.env.log_action(
            agent=self.settings.name,
            action="mobilization",
            reason=f"{self.settings.name} has arrived on site",
            additional="mobilization",
            duration=mobilization_hours,
            equipment_cost=self.settings.mobilization_cost,
        )

    def mobilize(self) -> Generator[Timeout, None, None]:
        """Mobilizes the ServiceEquipment object.

        NOTE: weather delays are not accounted for at this stage.

        Yields
        -------
        Generator[Timeout, None, None]
            A Timeout event for the number of hours the ServiceEquipment requires for
            mobilizing to the windfarm site.
        """
        self.env.log_action(
            agent=self.settings.name,
            action="mobilization",
            reason=f"{self.settings.name} is being mobilized",
            additional="mobilization",
        )

        self.enroute = True
        mobilization_hours = self.settings.mobilization_days * HOURS_IN_DAY
        yield self.env.timeout(mobilization_hours)
        self.onsite = True
        self.enroute = False

        self.env.log_action(
            agent=self.settings.name,
            action="mobilization",
            reason=f"{self.settings.name} has arrived on site",
            additional="mobilization",
            duration=mobilization_hours,
            equipment_cost=self.settings.mobilization_cost,
        )

    def _weather_forecast(self, hours: int | float) -> tuple[DatetimeIndex, np.ndarray]:
        """Retrieves the weather forecast from the simulation environment, and
        translates it to a boolean for satisfactory (True) and unsatisfactory (False)
        weather conditions.

        Parameters
        ----------
        hours : int | float
            The number of hours of weather data that should be retrieved.

        Returns
        -------
        tuple[DatetimeIndex, np.ndarray]
            The pandas DatetimeIndex and the boolean array of where the weather
            conditions are within safe operating limits for the servicing equipment
            (True) or not (False).
        """
        dt, wind, wave = self.env.weather_forecast(hours)
        all_clear = wind <= self.settings.max_windspeed_repair
        all_clear &= wave <= self.settings.max_waveheight_repair
        return dt, all_clear

    def _is_workshift(self, datetime_ix: DatetimeIndex) -> np.ndarray:
        """Determines which timestamps are in the servicing equipment's working hours.

        Parameters
        ----------
        datetime_ix : DatetimeIndex
            A pandas DatetimeIndex that needs to be assessed.

        Returns
        -------
        np.ndarray
            A boolean array for which values in working hours (True), and which values
            are outside working hours (False).
        """
        start = self.settings.workday_start
        end = self.settings.workday_end
        hour = datetime_ix.hour
        return (start <= hour) & (hour <= end)

    def find_uninterrupted_weather_window(
        self, hours_required: int | float
    ) -> tuple[int | float, bool]:
        """Finds the delay required before starting on a process that won't be able to
        be interrupted by a weather delay.

        TODO: WEATHER FORECAST NEEDS TO BE DONE FOR ``math.floor(self.now)``, not the ceiling
        or there will be a whole lot of rounding up errors on process times.

        Parameters
        ----------
        hours_required : int | float
            The number of uninterrupted of hours that a process requires for completion.

        Returns
        -------
        tuple[int | float, bool]
            The number of hours in weather delays before a process can be completed, and
            an indicator for if the process has to be delayed until the next shift for
            a safe transfer.
        """
        # If the hours required for the window is 0, then return 0 and indicate there is
        # no shift delay to be processed
        if hours_required == 0:
            return 0, False

        current = self.env.simulation_time

        # If the time required for a transfer is longer than the time left in the shift,
        # then return the hours left in the shift and indicate a shift delay
        max_hours = hours_until_future_hour(
            self.env.simulation_time, self.settings.workday_end
        )
        if hours_required > max_hours:
            return max_hours, True

        dt_ix, all_clear = self._weather_forecast(max_hours)
        all_clear &= self._is_workshift(dt_ix)
        safe_operating_windows = consecutive_groups(np.where(all_clear)[0])
        window_lengths = np.array([window.size for window in safe_operating_windows])

        # If all the safe operating windows are shorter than the time required, then
        # return the time until the end of the shift and indicate a shift delay
        if all([window < hours_required for window in window_lengths]):
            return max_hours, True

        # Find the length of the delay
        delay = safe_operating_windows[
            np.where(window_lengths >= hours_required)[0][0]
        ][0]

        # If there is no delay, simply return 0 and the shift delay indicator
        if delay == 0:
            return delay, False

        # If the delay is non-zero, ensure we get the correct hour difference from now
        # until the top of the first available hour
        delay = hours_until_future_hour(current, current.hour + delay)
        return delay, False

    def find_interrupted_weather_window(
        self, hours_required: int | float
    ) -> tuple[DatetimeIndex, np.ndarray, bool]:
        """Assesses at which points in the repair window, the wind (and wave)
        constraints for safe operation are met.

        The initial search looks for a weather window of length ``hours_required``, and
        adds 24 hours to the window for the proceeding 9 days. If no satisfactory window
        is found, then the calling process must make another call to this function, but
        likely there is something wrong with the constraints or weather conditions if
        no window is found within 10 days.

        Parameters
        ----------
        hours_required : int | float
            The number of hours required to complete the repair.

        Returns
        -------
        tuple[DatetimeIndex, np.ndarray, bool]
            The pandas DatetimeIndex, and a corresponding boolean array for what points
            in the time window are safe to complete a maintenance task or repair.
        """
        window_found = False
        hours_required = ceil(hours_required)

        for i in range(10):  # no more than 10 attempts to find a window
            dt_ix, weather = self._weather_forecast(hours_required + (i * 24))
            working_hours = self._is_workshift(dt_ix)
            window = weather & working_hours
            if window.sum() >= hours_required:
                window_found = True
                break

        return dt_ix, weather, window_found

    def get_next_request(self):
        """Gets the next request by the rig's method for processing repairs.

        Returns
        -------
        simpy.resources.store.FilterStoreGet
            The next ``RepairRequest`` to be processed.
        """
        if self.settings.method == "turbine":
            return self.manager.get_request_by_system(self.settings.capability)
        elif self.settings.method == "severity":
            return self.manager.get_next_highest_severity_request(
                self.settings.capability
            )

    def weather_delay(
        self, hours: int | float, **kwargs
    ) -> None | Generator[Timeout | Process, None, None]:
        """Processes a weather delay of length ``hours`` hours.

        Parameters
        ----------
        hours : int | float
            The lenght, in hours, of the weather delay.

        Returns
        -------
        None
            If the delay is 0 hours, then nothing happens.

        Yields
        -------
        None | Generator[Timeout | Process, None, None]
            If the delay is more than 0 hours, then a ``Timeout`` is yielded of length ``hours``.
        """
        if hours == 0:
            return

        salary_cost = self.calculate_salary_cost(hours)
        hourly_cost = self.calculate_hourly_cost(0)
        equipment_cost = self.calculate_equipment_cost(hours)
        self.env.log_action(
            duration=hours,
            action="delay",
            additional="weather delay",
            salary_labor_cost=salary_cost,
            hourly_labor_cost=hourly_cost,
            equipment_cost=equipment_cost,
            **kwargs,
        )
        yield self.env.timeout(hours)

    def repair(
        self,
        hours: int | float,
        request_details: Maintenance | Failure,
        **kwargs,
    ) -> None | Generator[Timeout | Process, None, None]:
        """[summary]

        Parameters
        ----------
        hours : int | float
            The lenght, in hours, of the repair or maintenance task.
        request_details : Maintenance | Failure
            The deatils of the request, this is only being checked for the type.

        Yields
        -------
        None | Generator[Timeout | Process, None, None]
            A ``Timeout`` is yielded of length ``hours``.
        """
        action = "repair" if isinstance(request_details, Failure) else "maintenance"
        salary_cost = self.calculate_salary_cost(hours)
        hourly_cost = self.calculate_hourly_cost(hours)
        equipment_cost = self.calculate_equipment_cost(hours)
        self.env.log_action(
            action=action,
            duration=hours,
            salary_labor_cost=salary_cost,
            hourly_labor_cost=hourly_cost,
            equipment_cost=equipment_cost,
            additional=action,
            **kwargs,
        )
        yield self.env.timeout(hours)

    @cache
    def _calculate_intra_site_time(self, start: str | None, end: str | None) -> float:
        """Calculates the time it takes to travel between port and site or between
        systems on site.

        Parameters
        ----------
        start : str | None
            The starting onsite location. If ``None``, then 0 is returned.
        end : str | None
            The ending onsite location. If ``None``, then 0 is returned.

        Returns
        -------
        float
            The travel time between two locations.
        """
        distance = 0  # setting for invalid cases to have no traveling
        travel_time = 0.0
        valid_sys = self.windfarm.distance_matrix.columns
        intra_site = start in valid_sys and end in valid_sys
        if intra_site:
            distance = self.windfarm.distance_matrix.loc[start, end]

        # if no speed is set, then there is no traveling time
        if self.settings.speed > 0:
            travel_time = distance / self.settings.speed

        # Infinity is the result of "traveling" between a system twice in a row
        if travel_time == float("inf"):
            travel_time = 0
        return travel_time

    def travel(
        self, start: str, end: str, set_current: str | None = None, **kwargs
    ) -> Generator[Timeout | Process, None, None]:
        """The process for traveling between port and site, or two systems onsite.

        NOTE: This does not currently take the weather conditions into account.

        Parameters
        ----------
        start : str
            The starting location, one of "site", "port", or "system".
        end : str
            The starting location, one of "site", "port", or "system".
        set_current : str, optional
            Where to set ``current_system`` to be if traveling to site or a different
            system onsite, by default None.

        Yields
        -------
        Generator[Timeout | Process, None, None]
            The timeout event for traveling.

        Raises
        ------
        ValueError
            Raised if ``start`` is not one of the required inputs.
        ValueError
            Raised if ``end`` is not one of the required inputs.
        """
        if start not in ("port", "site", "system"):
            raise ValueError(
                "``start`` location must be one of 'port', 'site', or 'system'!"
            )
        if end not in ("port", "site", "system"):
            raise ValueError(
                "``end`` location must be one of 'port', 'site', or 'system'!"
            )

        # TODO: Need a port-to-site distance to be implemented
        hours = 0  # takes care of invalid cases
        if start == end == "system":
            additional = f"traveling from {self.current_system} to {set_current}"
            hours = self._calculate_intra_site_time(self.current_system, set_current)
        else:
            additional = f"traveling from {start} to {end}"
            hours = 0

        # If the the equipment will arive after the shift is over, then it must travel
        # back to port (if needed), and wait for the next shift
        is_shift = self._is_workshift(self.env.simulation_time + timedelta(hours=hours))
        if not is_shift and end != "port" and not self.at_port:
            kw = {
                "additional": "insufficient time to complete travel before end of the shift"
            }
            kw.update(kwargs)
            yield self.env.process(self.travel(start=start, end="port", **kw))
            yield self.env.process(self.wait_until_next_shift(**kwargs))
            yield self.env.process(
                self.travel(start=start, end=end, set_current=set_current, **kwargs)
            )
        elif not is_shift and self.at_port:
            kw = {
                "additional": "insufficient time to complete travel before end of the shift"
            }
            kw.update(kwargs)
            yield self.env.process(self.wait_until_next_shift(**kwargs))
            yield self.env.process(
                self.travel(start=start, end=end, set_current=set_current, **kwargs)
            )

        salary_cost = self.calculate_salary_cost(hours)
        hourly_cost = self.calculate_hourly_cost(0)
        equipment_cost = self.calculate_equipment_cost(hours)
        self.env.log_action(
            action="traveling",
            additional=additional,
            duration=hours,
            salary_labor_cost=salary_cost,
            hourly_labor_cost=hourly_cost,
            equipment_cost=equipment_cost,
            **kwargs,
        )
        yield self.env.timeout(hours)

        if end == "port":
            self.at_port = True
            self.onsite = False
        else:
            self.at_port = False
            self.onsite = True
        self.at_system = True if set_current is not None else False
        self.current_system = set_current
        self.env.log_action(
            action="complete travel",
            additional=f"arrived at {set_current if set_current is not None else end}",
            **kwargs,
        )

    def crew_transfer(
        self,
        system: System,
        subassembly: Subassembly,
        request: RepairRequest,
        to_system: bool,
    ) -> None | Generator[Timeout | Process, None, None]:
        """The process of transfering the crew from the equipment to the ``System``
        for servicing using an uninterrupted weather window to ensure safe transfer.

        Parameters
        ----------
        system : System
            The System where the crew needs to be transferred to.
        subassembly : Subassembly
            The Subassembly that is being worked on.
        request : RepairRequest
            The repair to be processed.
        to_system : bool
            True if the crew is being transferred to the system, or False if the crew
            is being transferred off the system.

        Returns
        -------
        None
            None is returned when this is run recursively to not repeat the crew
            transfer process.

        Yields
        ------
        Generator[Timeout | Process, None, None]
            Yields a timeout event for the crew transfer once an interrupted weather
            window can be found.
        """
        hours_to_process = self.settings.crew_transfer_time
        salary_cost = self.calculate_salary_cost(hours_to_process)
        hourly_cost = self.calculate_hourly_cost(0)
        equipment_cost = self.calculate_equipment_cost(hours_to_process)

        shared_logging = dict(
            system_id=system.id,
            system_name=system.name,
            part_id=subassembly.id,
            part_name=subassembly.name,
            system_ol=system.operating_level,
            part_ol=subassembly.operating_level,
            agent=self.settings.name,
            reason=request.details.description,
            request_id=request.request_id,
        )

        delay, shift_delay = self.find_uninterrupted_weather_window(hours_to_process)
        # If there is a shift delay, then travel to port, wait, and travel back, and finally try again.
        if shift_delay:
            travel_time = self.env.now
            yield self.env.process(
                self.travel(start="site", end="port", **shared_logging)
            )
            travel_time -= self.env.now  # will be negative value, but is flipped
            delay -= abs(travel_time)  # decrement the delay by the travel time
            if delay > 0:
                yield self.env.process(self.weather_delay(delay, **shared_logging))
            yield self.env.process(
                self.wait_until_next_shift(
                    **shared_logging,
                    **{"additional": "weather unsuitable to transfer crew"},
                )
            )
            yield self.env.process(
                self.travel(
                    start="port", end="site", set_current=system.id, **shared_logging
                )
            )
            yield self.env.process(
                self.crew_transfer(system, subassembly, request, to_system=to_system)
            )
            return

        yield self.env.process(self.weather_delay(delay, **shared_logging))

        if to_system:
            additional = f"transferring crew from {self.settings.name} to {system.id}"
        else:
            additional = f"transferring crew from {system.id} to {self.settings.name}"
        self.env.log_action(
            action="transferring crew",
            additional=additional,
            duration=hours_to_process,
            salary_labor_cost=salary_cost,
            hourly_labor_cost=hourly_cost,
            equipment_cost=equipment_cost,
            **shared_logging,
        )
        self.transferring_crew = True
        yield self.env.timeout(hours_to_process)
        self.transferring_crew = False
        if to_system:
            self.current_system = system.id
        else:
            self.current_system = None
            self.at_system = False
        self.env.log_action(
            action="complete transfer", additional="complete", **shared_logging
        )

    def process_repair(
        self,
        request: RepairRequest,
        time_processed: int | float = 0,
        prior_operation_level: float = -1.0,
    ) -> Generator[Timeout | Process, None, None]:
        """Processes the repair including any weather and shift delays.

        Parameters
        ----------
        request : RepairRequest
            The ``Maintenance`` or ``Failure`` receiving attention.
        time_processed : int | float, optional
            Time that has already been processed, by default 0.
        prior_operation_level : float, optional
            The operating level of the ``System`` just before the repair has begun, by
            default -1.0.

        Yields
        -------
        Generator[Timeout | Process, None, None]
            Timeouts for the repair process.
        """

        """
        NOTE: THE PROCESS
        1. Travel to site/turbine, if not there already
        2. Transfer Crew
        3. Number of hours required, or hours until the end of the shift
        4. Do repairs for the above
        5. When shift/repair is complete, transfer crew
        6. Travel to next turbine, back to port

        7. Repeat until the repair is complete
        8. Exit this function
        """
        shift_delay = False

        if request.cable:
            system = self.windfarm.cable(request.system_id)
            cable = request.subassembly_id.split("::")[1:]
            subassembly = self.windfarm.graph.edges[cable]["cable"]
        else:
            system = self.windfarm.system(request.system_id)
            subassembly = getattr(system, request.subassembly_id)

        starting_operational_level = max(
            prior_operation_level, subassembly.operating_level
        )
        shared_logging = dict(
            system_id=system.id,
            system_name=system.name,
            part_id=subassembly.id,
            part_name=subassembly.name,
            system_ol=system.operating_level,
            part_ol=subassembly.operating_level,
            agent=self.settings.name,
            reason=request.details.description,
            request_id=request.request_id,
        )

        # Ensure there is enough time to transfer the crew back and forth with a buffer
        # of twice the travel time or travel back to port and try again the next shift
        start_shift = self.settings.workday_start
        end_shift = self.settings.workday_end
        current = self.env.simulation_time
        hours_required = request.details.time - time_processed
        hours_available = hours_until_future_hour(current, end_shift)

        if hours_available <= self.settings.crew_transfer_time * 4:
            yield self.env.process(
                self.travel(
                    start="site",
                    end="port",
                    **shared_logging,
                )
            )
            yield self.env.process(self.wait_until_next_shift(**shared_logging))

            yield self.env.process(
                self.process_repair(
                    request, prior_operation_level=starting_operational_level
                )
            )
            return

        # Travel to site or the next system on site
        if not self.at_system and self.at_port:
            yield self.env.process(
                self.travel(
                    start="port", end="site", set_current=system.id, **shared_logging
                )
            )
            # First turn off the turbine, then proceed with the servicing so the
            # turbine is not registered as operating when the turbine is being worked on
            system.servicing = True
            subassembly.interrupt_all_subassembly_processes()
            yield self.env.process(
                self.crew_transfer(system, subassembly, request, to_system=True)
            )
        elif self.at_system is not None and not self.at_port:
            yield self.env.process(
                self.travel(
                    start="system",
                    end="system",
                    set_current=system.id,
                    **shared_logging,
                )
            )
            yield self.env.process(
                self.crew_transfer(system, subassembly, request, to_system=True)
            )
        else:
            raise RuntimeError(f"{self.settings.name} is lost!")

        current = self.env.simulation_time

        # If the hours required is longer than the shift time, then reset the available
        # number of hours and the appropriate weather forecast, accounting for crew
        # transfer, otherwise get an adequate weather window, allowing for interruptions
        if not (start_shift == 0 and end_shift == 24):
            hours_available = hours_until_future_hour(current, end_shift)
            hours_available -= self.settings.crew_transfer_time
            _, weather_forecast = self._weather_forecast(hours_available)
        else:
            _, weather_forecast, _ = self.find_interrupted_weather_window(
                hours_required
            )

        # Check that all times are within the windfarm's working hours and cut off any
        # time points that fall outside of the work shifts
        if hours_required > hours_available:
            shift_delay = True

        hours_processed = 0
        weather_delay_groups = consecutive_groups(np.where(~weather_forecast)[0])
        while weather_forecast.size > 0 and hours_available > 0:
            delays = np.where(~weather_forecast)[0]
            if delays.size > 0:
                hours_to_process = delays[0]
                delay = weather_delay_groups.pop(0).size
            else:
                hours_to_process = weather_forecast.size
                delay = 0
            weather_forecast = weather_forecast[hours_to_process + delay :]

            # If the delay is at the start, hours_to_process is 0, and a delay gets
            # processed, otherwise the crew works for the minimum of
            # ``hours_to_process`` or maximum time that can be worked until the shift's
            # end, and maxed out by the hours required for the actual repair.
            if hours_to_process > 0:
                if hours_available <= hours_to_process:
                    hours_to_process = hours_available
                else:
                    current = self.env.simulation_time
                    hours_to_process = hours_until_future_hour(
                        current, current.hour + int(hours_to_process)
                    )
                if hours_required < hours_to_process:
                    hours_to_process = hours_required
                # Ensure this gets the correct float hours to the start of the target
                # hour, unless the hours to process is between (0, 1]
                yield self.env.process(
                    self.repair(hours_to_process, request.details, **shared_logging)
                )
                hours_processed += hours_to_process
                hours_available -= hours_to_process
                hours_required -= hours_to_process

            # If a delay is the first part of the process or a delay occurs after the
            # some work is performed, then that delay is processed here.
            if delay > 0 and hours_required > 0:
                current = self.env.simulation_time
                hours_to_process = hours_until_future_hour(
                    current, current.hour + delay
                )
                yield self.env.process(
                    self.weather_delay(hours_to_process, **shared_logging)
                )
                hours_available -= hours_to_process

        # Reached the end of the shift or the end of the repair, and need to unload crew
        # from the system
        yield self.env.process(
            self.crew_transfer(system, subassembly, request, to_system=False)
        )
        if shift_delay:
            yield self.env.process(
                self.travel(start="site", end="port", **shared_logging)
            )
            yield self.env.process(self.wait_until_next_shift(**shared_logging))

            yield self.env.process(
                self.process_repair(
                    request,
                    time_processed=hours_processed + time_processed,
                    prior_operation_level=starting_operational_level,
                )
            )
            return

        # Register the repair
        system.servicing = False
        self.register_repair_with_subassembly(
            subassembly, request, starting_operational_level
        )
        action = "maintenance" if isinstance(request.details, Maintenance) else "repair"
        self.env.log_action(
            system_id=system.id,
            part_id=subassembly.id,
            part_name=subassembly.name,
            system_ol=system.operating_level,
            part_ol=subassembly.operating_level,
            agent=self.settings.name,
            action=f"{action} complete",
            reason=request.details.description,
            materials_cost=request.details.materials,
            additional="complete",
            request_id=request.request_id,
        )

        # If this is the end of the shift, ensure that we're traveling back to port
        if not self.env.is_workshift(start_shift, end_shift):
            yield self.env.process(
                self.travel(start="site", end="port", **shared_logging)
            )

    def run_scheduled(self) -> Generator[Process, None, None]:
        """Runs the simulation.

        Yields
        -------
        Generator[Process, None, None]
            The simulation.
        """
        assert isinstance(self.settings, ScheduledServiceEquipmentData)  # mypy controls

        # If the starting operation date is the same as the simulations, set to be onsite
        if self.settings.operating_dates[0] == self.env.simulation_time.date():
            self.onsite = True

        while True:
            # Wait for a valid operational period to start
            if self.env.simulation_time.date() not in self.settings.operating_dates:  # type: ignore
                yield self.env.process(self.mobilize_scheduled())

            # Wait for next shift to start
            is_workshift = self.env.is_workshift(
                workday_start=self.settings.workday_start,
                workday_end=self.settings.workday_end,
            )
            if not is_workshift:
                yield self.env.process(
                    self.wait_until_next_shift(
                        **dict(
                            agent=self.settings.name,
                            reason="not in working hours",
                            additional="not in working hours",
                        )
                    )
                )

            request = self.get_next_request()
            if request is None:
                if not self.at_port:
                    yield self.env.process(
                        self.travel(
                            start="site",
                            end="port",
                            **dict(reason="no requests", agent=self.settings.name),
                        )
                    )
                yield self.env.process(
                    self.wait_until_next_shift(
                        **dict(
                            agent=self.settings.name,
                            reason="no requests",
                            additional="no work requests, waiting until the next shift",
                        )
                    )
                )
            else:
                yield self.env.process(self.process_repair(request.value))

    def run_unscheduled(self) -> Generator[Process, None, None]:
        """Runs an unscheduled maintenance simulation

        Yields
        -------
        Generator[Process, None, None]
            The simulation
        """

        charter_end_env_time = self.settings.charter_days * HOURS_IN_DAY  # type: ignore
        charter_end_env_time += self.settings.mobilization_days * HOURS_IN_DAY
        charter_end_env_time += self.env.now
        while True:
            if self.env.now >= charter_end_env_time:
                self.onsite = False
                self.at_port = False
                self.at_system = False
                self.current_system = None
                self.env.log_action(
                    agent=self.settings.name,
                    action="leaving site",
                    reason="charter period has ended",
                )
                break

            if not self.onsite:
                yield self.env.process(self.mobilize())

            # Wait for next shift to start
            is_workshift = self.env.is_workshift(
                workday_start=self.settings.workday_start,
                workday_end=self.settings.workday_end,
            )
            if not is_workshift:
                yield self.env.process(
                    self.wait_until_next_shift(
                        **dict(
                            agent=self.settings.name,
                            reason="not in working hours",
                            additional="not in working hours",
                        )
                    )
                )
            request = self.get_next_request()
            if request is None:
                yield self.env.process(
                    self.wait_until_next_shift(
                        **dict(
                            agent=self.settings.name,
                            reason="no requests",
                            additional="no work requests submitted by start of shift",
                        )
                    )
                )
            else:
                yield self.env.process(self.process_repair(request.value))



"""Creates the utilities objects."""


from .utilities import (
    IEC_power_curve,
    _mean,
    setup_logger,
    convert_dt_to_hours,
    hours_until_future_hour,
    format_events_log_message,
)



"""Creates the necessary repair classes."""


from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Sequence  # type: ignore
from itertools import chain
from collections import Counter

import numpy as np
from attrs import field, define
from simpy.resources.store import FilterStore, FilterStoreGet  # type: ignore

from wombat.core import Failure, Maintenance, RepairRequest, WombatEnvironment


if TYPE_CHECKING:
    from wombat.windfarm import Windfarm
    from wombat.core.service_equipment import ServiceEquipment


@define(auto_attribs=True)
class EquipmentMap:
    """Internal mapping for servicing equipment strategy information."""

    strategy_threshold: int | float
    equipment: ServiceEquipment


@define(auto_attribs=True)
class StrategyMap:
    """Internal mapping for equipment capabilities and their data."""

    CTV: list[EquipmentMap] = field(factory=list)
    SCN: list[EquipmentMap] = field(factory=list)
    LCN: list[EquipmentMap] = field(factory=list)
    CAB: list[EquipmentMap] = field(factory=list)
    RMT: list[EquipmentMap] = field(factory=list)
    DRN: list[EquipmentMap] = field(factory=list)
    DSV: list[EquipmentMap] = field(factory=list)
    is_running: bool = field(default=False, init=False)

    def update(
        self, capability: str, threshold: int | float, equipment: ServiceEquipment
    ) -> None:
        """A method to update the strategy mapping between capability types and the
        available ``ServiceEquipment`` objects.

        Parameters
        ----------
        capability : str
            The ``equipment``'s capability.
        threshold : int | float
            The threshold for ``equipment``'s strategy.
        equipment : ServiceEquipment
            The actual ``ServiceEquipment`` object to be logged.

        Raises
        ------
        ValueError
            Raised if there is an invalid capability, though this shouldn't be able to
            be reached.
        """
        # Using a mypy ignore because of an unpatched bug using data classes
        if capability == "CTV":
            self.CTV.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "SCN":
            self.SCN.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "LCN":
            self.LCN.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "CAB":
            self.CAB.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "RMT":
            self.RMT.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "DRN":
            self.DRN.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "DSV":
            self.DSV.append(EquipmentMap(threshold, equipment))  # type: ignore
        else:
            # This should not even be able to be reached
            raise ValueError(
                f"Invalid servicing equipment '{capability}' has been provided!"
            )
        self.is_running = True


class RepairManager(FilterStore):
    """Provides a class to manage repair and maintenance tasks.

    Parameters
    ----------
    FilterStore : simpy.resources.store.FilterStore
        The ``simpy`` class on which RepairManager is based to manage the repair and
        maintenance tasks.
    env : wombat.core.WombatEnvironment
        The simulation environment.
    capacity : float
        The maximum number of tasks that can be submitted to the manager, by default ``np.inf``.

    Attributes
    ----------
    env : wombat.core.WombatEnvironment
        The simulation environment.
    windfarm: wombat.windfarm.Windfarm
        The simulated windfarm. This is only used for getting the operational capacity.
    _current_id : int
        The logged and auto-incrememented integer base for the ID generated for each
        submitted repair request.
    downtime_based_equipment: StrategyMap
        The mapping between downtime-based servicing equipment and their capabilities.
    request_based_equipment: StrategyMap
        The mapping between request-based servicing equipment and their capabilities.
    """

    def __init__(self, env: WombatEnvironment, capacity: float = np.inf) -> None:
        super().__init__(env, capacity)

        self.env = env
        self._current_id = 0

        self.downtime_based_equipment = StrategyMap()
        self.request_based_equipment = StrategyMap()

    def _update_equipment_map(self, service_equipment: ServiceEquipment) -> None:
        """Updates ``equipment_map`` with a provided servicing equipment object."""
        capability = service_equipment.settings.capability
        strategy = service_equipment.settings.strategy

        if strategy == "downtime":
            mapping = self.downtime_based_equipment
        elif strategy == "requests":
            mapping = self.request_based_equipment
        else:
            # Shouldn't be possible to get here!
            raise ValueError("Invalid servicing equipment!")

        strategy_threshold = service_equipment.settings.strategy_threshold  # type: ignore
        if isinstance(capability, list):
            for c in capability:
                mapping.update(c, strategy_threshold, service_equipment)

    def _register_windfarm(self, windfarm: Windfarm) -> None:
        """Adds the simulation windfarm to the class attributes"""
        self.windfarm = windfarm

    def _register_equipment(self, service_equipment: ServiceEquipment) -> None:
        """Adds the servicing equipment to the class attributes and adds it to the
        capabilities mapping.
        """
        self._update_equipment_map(service_equipment)

    def _create_request_id(self, request: RepairRequest) -> str:
        """Creates a unique ``request_id`` to be logged in the ``request``.

        Parameters
        ----------
        request : RepairRequest
            The request object.

        Returns
        -------
        str
            An 11-digit identifier starting with "MNT" for maintenance tasks or "RPR"
            for repairs.

        Raises
        ------
        ValueError
            If the ``request.details`` property is not a ``Failure`` or ``Maintenance`` object,
            then a ValueError will be raised.
        """
        if isinstance(request.details, Failure):
            prefix = "RPR"
        elif isinstance(request.details, Maintenance):
            prefix = "MNT"
        else:
            # Note this is a safety, and shouldn't be able to be reached
            raise ValueError("A valid ``RepairRequest`` must be submitted")

        request_id = f"{prefix}{str(self._current_id).zfill(8)}"
        self._current_id += 1
        return request_id

    def _run_equipment_downtime(self) -> None:
        """Run any equipment that has a pending request where the current windfarm
        operating capacity is less than or equal to the servicing equipment's threshold.
        """
        operating_capacity = self.windfarm.current_availability_wo_servicing
        for capability in self.request_map:
            equipment_mapping = getattr(self.downtime_based_equipment, capability)
            for equipment in equipment_mapping:
                if operating_capacity > equipment.strategy_threshold:
                    continue
                if equipment.equipment.onsite or equipment.equipment.enroute:
                    continue
                self.env.process(equipment.equipment.run_unscheduled())

    def _run_equipment_requests(self) -> None:
        """Run the first piece of equipment (if none are onsite) for each equipment
        capability category where the number of requests is greater than or equal to the
        equipment's threshold.
        """
        for capability, n_requests in self.request_map.items():
            equipment_mapping = getattr(self.request_based_equipment, capability)
            for i, equipment in enumerate(equipment_mapping):
                if n_requests < equipment.strategy_threshold:
                    continue
                # Run only the first piece of equipment in the mapping list, but ensure
                # that it moves to the back of the line after being used
                if equipment.equipment.onsite or equipment.equipment.enroute:
                    equipment_mapping.append(equipment_mapping.pop(i))
                    break
                self.env.process(equipment.equipment.run_unscheduled())
                equipment_mapping.append(equipment_mapping.pop(i))

    def submit_request(self, request: RepairRequest) -> RepairRequest:
        """The method to submit requests to the repair mananger and adds a unique
        identifier for logging.

        Parameters
        ----------
        request : RepairRequest
            The request that needs to be tracked and logged.

        Returns
        -------
        RepairRequest
            The same request as passed into the method, but with a unique identifier
            used for logging.
        """
        request_id = self._create_request_id(request)
        request.assign_id(request_id)
        self.put(request)

        if self.downtime_based_equipment.is_running:
            self._run_equipment_downtime()
        if self.request_based_equipment.is_running:
            self._run_equipment_requests()

        return request

    def get_request_by_system(
        self, equipment_capability: Sequence[str], system_id: Optional[str] = None
    ) -> Optional[FilterStoreGet]:
        """Gets all repair requests for a certain turbine with given a sequence of
        ``equipment_capability``.

        Parameters
        ----------
        equipment_capability : list[str]
            The capability of the servicing equipment requesting repairs to process.
        system_id : Optional[str], optional
            ID of the turbine or OSS; should correspond to ``System.id``, by default
            None. If None, then it will simply sort the list by ``System.id`` and return
            the first repair requested.

        Returns
        -------
        Optional[FilterStoreGet]
            The first repair request for the focal system or None, if self.items is
            empty, or there is no matching system.
        """
        if not self.items:
            return None

        equipment_capability = set(equipment_capability)  # type: ignore

        # Filter the requests by system
        requests = self.items
        if system_id is not None:
            requests = [el for el in self.items if el.system_id == system_id]
        if requests == []:
            return None

        # Filter the requests by equipment capability and return the first valid request
        for request in requests:
            if equipment_capability.intersection(request.details.service_equipment):  # type: ignore
                return self.get(lambda x: x == requests[0])

        # In case the loop above iterates all the way to the end, nothing was
        # found so return None. This is probably an error for which an error
        # should be raised, but this keeps the type hints correct for now.

        return None

    def get_next_highest_severity_request(
        self, equipment_capability: Sequence[str], severity_level: Optional[int] = None
    ) -> Optional[FilterStoreGet]:
        """Gets the next repair request by ``severity_level``.

        Parameters
        ----------
        equipment_capability : list[str]
            The capability of the equipment requesting possible repairs to make.
        severity_level : int
            Severity level of focus, default None.

        Returns
        -------
        Optional[FilterStoreGet]
            Repair request meeting the required criteria.
        """
        if not self.items:
            return None

        equipment_capability = set(equipment_capability)  # type: ignore

        # Filter the requests by severity level
        requests = self.items
        if severity_level is not None:
            requests = [el for el in self.items if el.severity_level == severity_level]
        if requests == []:
            return None

        requests = sorted(requests, key=lambda x: x.severity_level, reverse=True)
        for request in requests:
            if equipment_capability.intersection(request.details.service_equipment):  # type: ignore
                return self.get(lambda x: x == request)

        # This return statement ensures there is always a known return type,
        # but consider raising an error here if the loop is always assumed to
        # break when something is found.

        return None

    def purge_subassembly_requests(
        self, system_id: str, subassembly_id: str, exclude: list[str] = []
    ) -> Optional[list[RepairRequest]]:
        """Yields all the requests for a system/subassembly combination. This is intended
        to be used to remove erroneous requests after a subassembly has been replaced.

        Parameters
        ----------
        system_id : str
            ``System.id`` or ``Cable.id``.
        subassembly_id : str
            ``Subassembly.id`` or the ``Cable.id`` repeated for cables.
        exclude : list[str]
            A list of ``request_id``s to exclude from the purge. This is a specific use
            case for the combined cable system/subassembly, but can be to exclude
            certain requests from the purge.

        Yields
        -------
        Optional[list[RepairRequest]]
            All requests made to the repair manager for the provided system/subassembly
            combination. Returns None if self.items is empty or the loop terminates
            without finding what it is looking for.
        """
        if not self.items:
            return None

        requests = [
            request
            for request in self.items
            if (
                request.system_id == system_id
                and request.subassembly_id == subassembly_id
                and request.request_id not in exclude
            )
        ]
        if requests == []:
            return None

        for request in requests:
            self.env.log_action(
                system_id=request.system_id,
                system_name=request.system_name,
                part_id=request.subassembly_id,
                part_name=request.subassembly_name,
                system_ol=float("nan"),
                part_ol=float("nan"),
                agent="RepairManager",
                action="request canceled",
                reason="replacement required",
                request_id=request.request_id,
            )
            _ = self.get(lambda x: x == request)  # pylint: disable=W0640
        return requests

    @property
    def request_map(self) -> dict[str, int]:
        """Creates an updated mapping between the servicing equipment capabilities and
        the number of requests that fall into each capability category (nonzero values only).
        """
        # mapping = dict(CTV=0, SCN=0, LCN=0, CAB=0, RMT=0, DRN=0, DSV=0)
        all_requests = [request.details.service_equipment for request in self.items]
        requests = dict(Counter(chain.from_iterable(all_requests)))
        # mapping.update(requests)
        return requests



"""Provides the O&M Enviroment class; a subclass of simpy.Environment."""
from __future__ import annotations

import math  # type: ignore
import logging  # type: ignore
import datetime as dt  # type: ignore
from typing import Tuple, Union, Optional  # type: ignore
from pathlib import Path  # type: ignore
from datetime import datetime, timedelta  # type: ignore

import numpy as np  # type: ignore
import simpy  # type: ignore
import pandas as pd  # type: ignore
from simpy.events import Event  # type: ignore
from pandas.core.indexes.datetimes import DatetimeIndex

import wombat  # pylint: disable=W0611
from wombat.utilities import (
    setup_logger,
    hours_until_future_hour,
    format_events_log_message,
)


class WombatEnvironment(simpy.Environment):
    """The primary mechanism for powering an O&M simulation. This object has insight into
    all other simulation objects, and controls the timing, date/time stamps, and weather
    conditions.
    """

    def __init__(
        self,
        data_dir: Path | str,
        weather_file: str,
        workday_start: int,
        workday_end: int,
        simulation_name: str | None = None,
        start_year: int | None = None,
        end_year: int | None = None,
    ) -> None:
        """Initialization.

        Parameters
        ----------
        data_dir : pathlib.Path | str
            Directory where the inputs are stored and where to save outputs.
        weather_file : str
            Name of the weather file. Should be contained within ``data_dir``/weather/.
        workday_start : int
            Starting time for the repair crew, in 24 hour local time. This can be
            overridden by an ``ServiceEquipmentData`` object that operates outside of the "typical"
            working hours.
        workday_end : int
            Ending time for the repair crew, in 24 hour local time. This can be
            overridden by an ``ServiceEquipmentData`` object that operates outside of the "typical"
            working hours.
        simulation_name : str | None, optional
            Name of the simulation; will be used for naming the log file, by default ``None``.
            If ``None``, then the current time will be used. Will always save to
            ``data_dir``/outputs/logs/``simulation_name``.log.
            ... note: spaces (" ") will be replaced with underscores ("_"), for example:
            "my example analysis" becomes "my_example_analysis".
        start_year : int | None, optional
            Custom starting year for the weather profile, by default None. If ``None`` or
            less than the first year of the weather profile, this will be ignored.
        end_year : int | None, optional
            Custom ending year for the weather profile, by default None. If ``None`` or
            greater than the last year of the weather profile, this will be ignored.

        Raises
        ------
        FileNotFoundError
            Raised if ``data_dir`` cannot be found.
        """
        super().__init__()
        self.data_dir = Path(data_dir).resolve()
        if not self.data_dir.is_dir():
            raise FileNotFoundError(f"{self.data_dir} does not exist")

        self.workday_start = int(workday_start)
        self.workday_end = int(workday_end)
        if not 0 <= self.workday_start < 24:
            raise ValueError("workday_start must be a valid 24hr time before midnight!")
        if not 0 <= self.workday_end < 24:
            raise ValueError("workday_end must be a valid 24hr time before midnight!")
        if self.workday_end <= self.workday_start:
            raise ValueError(
                "Work shifts must end after they start ({self.workday_start}hrs)."
            )

        self.weather = self._weather_setup(weather_file, start_year, end_year)
        self.max_run_time = self.weather.shape[0]
        self.shift_length = self.workday_end - self.workday_start

        self.simulation_name = simulation_name
        self._logging_setup()

    def run(self, until: Optional[Union[int, float, Event]] = None):
        """Extends the ``simpy.Environment.run`` method to change the default behavior if
        no argument is passed to ``until``, which will now run a simulation until the end
        of the weather profile is reached.

        Parameters
        ----------
        until : Optional[Union[int, float, Event]], optional
            When to stop the simulation, by default None. See documentation on
            ``simpy.Environment.run`` for more details.
        """
        if until is None:
            until = self.max_run_time
        elif until > self.max_run_time:
            until = self.max_run_time
        super().run(until=until)

    def _logging_setup(self) -> None:
        """Completes the setup for logging data."""
        if self.simulation_name is None:
            self.simulation_name = simulation = "wombat"
        else:
            simulation = self.simulation_name.replace(" ", "_")
        dt_stamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S.%f")

        events_log_fname = f"{dt_stamp}_{simulation}_events.log"
        operations_log_fname = f"{dt_stamp}_{simulation}_operations.log"
        power_potential_fname = f"{dt_stamp}_{simulation}_power_potential.csv"
        power_production_fname = f"{dt_stamp}_{simulation}_power_production.csv"
        metrics_input_fname = f"{dt_stamp}_{simulation}_metrics_inputs.yaml"

        log_path = self.data_dir / "outputs" / "logs"
        self.events_log_fname = log_path / events_log_fname
        self.operations_log_fname = log_path / operations_log_fname
        self.power_potential_fname = log_path / power_potential_fname
        self.power_production_fname = log_path / power_production_fname
        self.metrics_input_fname = log_path / metrics_input_fname

        _dir = self.data_dir / "outputs" / "logs"
        if not _dir.is_dir():
            _dir.mkdir()

        setup_logger("events_log", self.events_log_fname)
        setup_logger("operations_log", self.operations_log_fname)
        self._events_logger = logging.getLogger("events_log")
        self._operations_logger = logging.getLogger("operations_log")

    @property
    def simulation_time(self) -> datetime:
        """Current time within the simulation ("datetime" column within weather)."""
        now = self.now
        minutes = now % 1 * 60
        dt = self.weather.iloc[math.floor(now)].name.to_pydatetime()
        minutes, seconds = math.floor(minutes), math.ceil(minutes % 1 * 60)
        return dt + timedelta(minutes=minutes, seconds=seconds)

    def is_workshift(self, workday_start: int = -1, workday_end: int = -1) -> bool:
        """Checks if the current simulation time is within the windfarm's working hours.

        Parameters
        ----------
        workday_start : int
            A valid hour in 24 hour time, by default -1. This should only be provided from an
            ``ServiceEquipmentData`` object. ``workday_end`` must also be provided in order to be used.
        workday_end : int
            A valid hour in 24 hour time, by default -1. This should only be provided from an
            ``ServiceEquipmentData`` object. ``workday_start`` must also be provided in order to be used.

        Returns
        -------
        bool
            True if it's valid working hours, False otherwise.
        """
        if -1 in (workday_start, workday_end):
            # Return true if the shift is around the clock
            if self.workday_start == 0 and self.workday_end == 24:
                return True
            return self.workday_start <= self.simulation_time.hour < self.workday_end

        # Return true if the shift is around the clock
        if workday_start == 0 and workday_end == 24:
            return True

        return workday_start <= self.simulation_time.hour < workday_end

    def hour_in_shift(
        self, hour: int, workday_start: int = -1, workday_end: int = -1
    ) -> bool:
        """Checks whether an ``hour`` is within the working hours.

        Parameters
        ----------
        hour : int
            Hour of the day.
        workday_start : int
            A valid hour in 24 hour time, by default -1. This should only be provided from an
            ``ServiceEquipmentData`` object. ``workday_end`` must also be provided in order to be used.
        workday_end : int
            A valid hour in 24 hour time, by default -1. This should only be provided from an
            ``ServiceEquipmentData`` object. ``workday_start`` must also be provided in order to be used.

        Returns
        -------
        bool
            True if ``hour`` is during working hours, False otherwise.
        """
        if -1 in (workday_start, workday_end):
            return self.workday_start <= hour < self.workday_end
        return workday_start <= hour < workday_end

    def hours_to_next_shift(self, workday_start: int = -1) -> float:
        """Time until the next work shift starts, in hours.

        Parameters
        ----------
        workday_start : int
            A valid hour in 24 hour time, by default -1. This should only be provided
            from an ``ServiceEquipmentData`` object.

        Returns
        -------
        float
            Hours until the next shift starts.
        """
        current = self.simulation_time
        start = self.workday_start if workday_start == -1 else workday_start
        if current.hour < start:
            # difference between now and workday start
            return hours_until_future_hour(current, start)
        elif current.hour == start == 0:
            # Need to manually move forward one whole day to avoid an infinite loop
            return hours_until_future_hour(current, 24)
        else:
            # time to midnight + hour of workday start
            return start + hours_until_future_hour(current, 0)

    @property
    def current_time(self) -> str:
        """Timestamp for the current time as a datetime.datetime.strftime."""
        return datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")

    def date_ix(self, date: dt.datetime | dt.date) -> int:
        """The first index of a future date. This corresponds to the number of hours
        until this dates from the very beginning of the simulation.

        Parameters
        ----------
        date : datetime.datetime | datetime.date
            A date within the environment's simulation range.

        Returns
        -------
        int
            Index of the weather profile corresponds to the first hour of ``date``.
        """
        if isinstance(date, dt.datetime):
            date = date.date()
        return np.where(self.weather.index.date == date)[0][0]

    def _weather_setup(
        self,
        weather_file: str,
        start_year: Optional[int] = None,
        end_year: Optional[int] = None,
    ) -> pd.DataFrame:
        """Reads the weather data from the "<inputs>/weather" directory, and creates the
        ``start_date`` and ``end_date`` time stamps for the simulation.

        This also fills any missing data with zeros and interpolates the values of any
        missing datetime entries.

        Parameters
        ----------
        weather_file : str
            Name of the weather file to be used by the environment. Should be contained
            within ``data_dir/weather``.
        start_year : Optional[int], optional
            Custom starting year for the weather profile, by default None. If ``None``
            or less than the first year of the weather profile, this will be ignored.
        end_year : Optional[int], optional
            Custom ending year for the weather profile, by default None. If ``None`` or
            greater than the last year of the weather profile, this will be ignored.

        Returns
        -------
        pd.DataFrame
            The wind (and  wave) timeseries.
        """
        weather_path = self.data_dir / "weather" / weather_file
        weather = pd.read_csv(
            weather_path, parse_dates=["datetime"], index_col="datetime"
        )
        weather = weather.fillna(0.0)
        weather = weather.resample("H").interpolate(limit_direction="both", limit=5)

        # Create the start and end points
        self.start_datetime = weather.index[0].to_pydatetime()
        self.end_datetime = weather.index[-1].to_pydatetime()
        self.start_year = self.start_datetime.year
        self.end_year = self.end_datetime.year

        if start_year is None and end_year is None:
            return weather

        if start_year is None:
            pass
        elif start_year > self.end_year:
            raise ValueError(
                f"'stary_year' ({start_year}) occurs after the last available year"
                f" in the weather data (range: {self.end_year})"
            )
        else:
            # Filter for the provided, validated starting year and update the attribute
            weather = weather[weather.index.year >= start_year]
            self.start_datetime = weather.index[0].to_pydatetime()
            start_year = self.start_year = self.start_datetime.year

        if end_year is None:
            pass
        elif start_year is None and end_year < self.start_year:
            raise ValueError(
                f"The provided 'end_year' ({end_year}) is before the start_year"
                f" ({self.start_year})"
            )
        elif start_year is not None:
            if end_year < start_year:
                raise ValueError(
                    f"The provided 'end_year' ({end_year}) is before the start_year"
                    f" ({start_year})"
                )
            else:
                # Filter for the provided, validated ending year and update the attribute
                weather = weather[weather.index.year <= end_year]
                self.end_datetime = weather.index[-1].to_pydatetime()
                self.end_year = self.end_datetime.year
        else:
            # Filter for the provided, validated ending year and update the attribute
            weather = weather[weather.index.year <= end_year]
            self.end_datetime = weather.index[-1].to_pydatetime()
            self.end_year = self.end_datetime.year

        return weather

    @property
    def weather_now(self) -> Tuple[float, float]:
        """The current weather.

        Returns
        -------
        Tuple[float, float]
            Wind and wave data for the current time.
        """
        # Rounds down because we won't arrive at the next weather event until that hour
        now = int(self.now)
        return self.weather.iloc[now].values

    def weather_forecast(
        self, hours: Union[int, float]
    ) -> Tuple[DatetimeIndex, np.ndarray, np.ndarray]:
        """Returns the wind and wave data for the next ``hours`` hours, starting from
        the current hour's weather.

        Parameters
        ----------
        hours : Union[int, float]
            Number of hours to look ahead, rounds up to the nearest hour.

        Returns
        -------
        Tuple[DatetimeIndex, np.ndarray, np.ndarray]
            The pandas DatetimeIndex, windspeed array, and waveheight array for the
            hours requested, each with shape (``hours`` + 1).
        """
        start = math.floor(self.now)
        end = start + math.ceil(hours)

        # If it's not on the hour, ensure we're looking ``hours`` hours into the future
        if self.now % 1:
            end += 1

        weather = self.weather.iloc[start:end]
        return (weather.index, weather.windspeed.values, weather.waveheight.values)

    def log_action(
        self,
        *,
        agent: str,
        action: str,
        reason: str,
        additional: str = "",
        system_id: str = "",
        system_name: str = "",
        part_id: str = "",
        part_name: str = "",
        system_ol: float | int = 0,
        part_ol: float | int = 0,
        duration: float = 0,
        request_id: str = "na",
        materials_cost: Union[int, float] = 0,
        hourly_labor_cost: Union[int, float] = 0,
        salary_labor_cost: Union[int, float] = 0,
        equipment_cost: Union[int, float] = 0,
    ) -> None:
        """Formats the logging messages into the expected format for logging.

        Parameters
        ----------
        agent : str
            Agent performing the action.
        action : str
            Action that was taken.
        reason : str
            Reason an action was taken.
        additional : str
            Any additional information that needs to be logged.
        system_id : str
            Turbine ID, ``System.id``, by default "".
        system_name : str
            Turbine name, ``System.name``, by default "".
        part_id : str
            Subassembly, component, or cable ID, ``_.id``, by default "".
        part_name : str
            Subassembly, component, or cable name, ``_.name``, by default "".
        system_ol : float | int
            Turbine operating level, ``System.operating_level``. Use an empty string for n/a, by default 0.
        part_ol : float | int
            Subassembly, component, or cable operating level, ``_.operating_level``. Use
            an empty string for n/a, by default 0.
        request_id : str
            The ``RepairManager`` assigned request_id found in ``RepairRequest.request_id``, by default "na".
        duration : float
            Length of time the action lasted, by default 0.
        materials_cost : Union[int, float], optional
            Total cost of materials for action, in USD, by default 0.
        hourly_labor_cost : Union[int, float], optional
            Total cost of hourly labor for action, in USD, by default 0.
        salary_labor_cost : Union[int, float], optional
            Total cost of salaried labor for action, in USD, by default 0.
        equipment_cost : Union[int, float], optional
            Total cost of equipment for action, in USD, by default 0.
        """
        self._events_logger.info(
            format_events_log_message(
                self.simulation_time,
                self.now,
                system_id,
                system_name,
                part_id,
                part_name,
                system_ol,
                part_ol,
                agent,
                action,
                reason,
                additional,
                duration,
                request_id,
                materials_cost,
                hourly_labor_cost,
                salary_labor_cost,
                equipment_cost,
            )
        )

    def _create_events_log_dataframe(self) -> pd.DataFrame:
        """Imports the logging file created in ``run`` and returns it as a formatted
        ``pandas.DataFrame``.

        Returns
        -------
        pd.DataFrame
            The formatted logging data from a simulation.
        """
        log_columns = [
            "datetime",
            "name",
            "level",
            "env_datetime",
            "env_time",
            "system_id",
            "system_name",
            "part_id",
            "part_name",
            "system_operating_level",
            "part_operating_level",
            "agent",
            "action",
            "reason",
            "additional",
            "duration",
            "request_id",
            "materials_cost",
            "hourly_labor_cost",
            "salary_labor_cost",
            "equipment_cost",
            "total_labor_cost",
            "total_cost",
        ]
        log_df = pd.read_csv(
            self.events_log_fname,
            names=log_columns,
            sep=" :: ",
            index_col=False,
            engine="python",
        )
        log_df = log_df.loc[log_df.level == "INFO"]
        log_df = log_df.drop(labels=["name", "level"], axis=1)
        log_df.datetime = pd.to_datetime(log_df.datetime)
        log_df.env_datetime = pd.to_datetime(log_df.env_datetime)
        log_df.index = log_df.datetime
        log_df = log_df.drop(labels="datetime", axis=1)
        return log_df

    def _create_operations_log_dataframe(self) -> pd.DataFrame:
        """Imports the logging file created in ``run`` and returns it as a formatted
        ``pandas.DataFrame``.

        Returns
        -------
        pd.DataFrame
            The formatted logging data from a simulation.
        """
        with open(self.operations_log_fname) as f:
            first_line = f.readline().strip().split(" :: ")
            column_names = first_line[3:]

        log_df = pd.read_csv(
            self.operations_log_fname,
            names=column_names,
            skiprows=1,
            sep=" :: ",
            index_col=False,
            engine="python",
        )

        log_df = log_df.loc[log_df.level == "INFO"]
        log_df = log_df.drop(labels=["name", "level"], axis=1)
        log_df.datetime = pd.to_datetime(log_df.datetime)
        log_df.env_datetime = pd.to_datetime(log_df.env_datetime)
        log_df.index = log_df.datetime
        log_df = log_df.drop(labels="datetime", axis=1)
        return log_df

    def convert_logs_to_csv(  # type: ignore
        self, delete_original: bool = False, return_df: bool = True
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Creates a CSV file for the both the events and operations logs.

        Parameters
        ----------
        delete_original : bool, optional
            If True, delete the corresponding ".log" files, by default False.
        return_df : bool, optional
            If True, returns the pd.DataFrame objects for the operations and events
            logging, by default True.

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame]
            Returns nothing if ``return_df`` is False, otherwise returns the operations and
            events log ``pd.DataFrame`` objects in that order.
        """
        events = self._create_events_log_dataframe().sort_values("env_time")
        operations = self._create_operations_log_dataframe().sort_values("env_time")

        events_fname = self.events_log_fname.with_suffix(".csv")
        events.to_csv(events_fname, index=False)

        operations_fname = self.operations_log_fname.with_suffix(".csv")
        operations.to_csv(operations_fname, index=False)

        if delete_original:
            self.operations_log_fname.unlink()
            self.events_log_fname.unlink()

        if return_df:
            return operations, events

    def power_production_potential_to_csv(  # type: ignore
        self,
        windfarm: "wombat.windfarm.Windfarm",  # type: ignore
        operations: Optional[pd.DataFrame] = None,
        return_df: bool = True,
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Creates the power production ``DataFrame`` and optionally returns it.

        Parameters
        ----------
        windfarm : wombat.windfarm.Windfarm
            The simulation's windfarm object.
        operations : Optional[pd.DataFrame], optional
            The operations log ``DataFrame`` if readily available, by default None. If
            ``None``, then it will be created through ``convert_logs_to_csv()``.
        return_df : bool, optional
            Indicator to return the power production for further usage, by default True.

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame]
            The power potential and production timeseries data.
        """
        if operations is None:
            operations = self._create_operations_log_dataframe().sort_values("env_time")

        turbines = windfarm.turbine_id
        windspeed = self.weather.windspeed
        windspeed = windspeed.loc[operations.env_datetime]

        potential = np.vstack(
            ([windfarm.system(t_id).power(windspeed) for t_id in turbines])
        ).T
        potential_df = pd.DataFrame(
            [],
            index=windspeed.index,
            columns=["env_time", "env_datetime", "windspeed", "windfarm"]
            + turbines.tolist(),
        )
        potential_df[turbines] = potential
        potential_df.windspeed = windspeed.values
        potential_df.windfarm = potential.sum(axis=1)
        potential_df.env_time = operations.env_time.values
        potential_df.env_datetime = operations.env_datetime.values
        potential_df.to_csv(self.power_potential_fname, index_label="env_datetime")

        production_df = potential_df.copy()
        production_df[turbines] = (
            production_df[turbines].values * operations[turbines].values
        )
        production_df.windfarm = production_df[turbines].sum(axis=1)
        production_df.to_csv(self.power_production_fname, index_label="env_datetime")
        if return_df:
            return potential_df, production_df

    def cleanup_log_files(self, log_only=False) -> None:
        """This is a convenience method to clear the output log files in case a large
        batch of simulations is being run and there are space limitations.

        Parameters
        ----------
        log_only : bool, optional
            Only deletes the xx.log files, if True, otherwise the xx.log and xx.csv
            logging files are all deleted, by default False
        """

        # NOTE: Everything is wrapped in a try/except clause to protect against failure
        # when inevitably a file has already been deleted on accident, or if in the
        # dataframe generation step, the original logs were deleted

        try:
            self.events_log_fname.unlink()
        except FileNotFoundError:
            pass

        try:
            self.operations_log_fname.unlink()
        except FileNotFoundError:
            pass

        if not log_only:
            # Don't fail if any of the following files were not created
            try:
                self.events_log_fname.with_suffix(".csv").unlink()
            except FileNotFoundError:
                pass

            try:
                self.operations_log_fname.with_suffix(".csv").unlink()
            except FileNotFoundError:
                pass

            try:
                self.power_potential_fname.unlink()
            except FileNotFoundError:
                pass

            try:
                self.power_production_fname.unlink()
            except FileNotFoundError:
                pass

            try:
                self.metrics_input_fname.unlink()
            except FileNotFoundError:
                pass



"""Creates the Turbine class."""
from __future__ import annotations

from typing import Callable  # type: ignore
from functools import reduce

import numpy as np
import pandas as pd

from wombat.core import RepairManager, WombatEnvironment
from wombat.utilities import IEC_power_curve
from wombat.windfarm.system import Subassembly


try:  # pylint: disable=duplicate-code
    from functools import cache  # type: ignore
except ImportError:  # pylint: disable=duplicate-code
    from functools import lru_cache  # pylint: disable=duplicate-code

    cache = lru_cache(None)  # pylint: disable=duplicate-code


@cache
def _product(x: float, y: float) -> float:
    """Multiplies two numbers. Used for a reduce operation.

    Parameters
    ----------
    x : float
        The first number.
    y : float
        The second number.

    Returns
    -------
    float
        The product of the two numbers.
    """
    return x * y


class System:
    """Can either be a turbine or substation, but is meant to be something that consists
    of 'Subassembly' pieces.

    See `here <https://www.sciencedirect.com/science/article/pii/S1364032117308985>`_
    for more information.
    """

    def __init__(
        self,
        env: WombatEnvironment,
        repair_manager: RepairManager,
        t_id: str,
        name: str,
        subassemblies: dict,
        system: str,
    ):
        """Initializes an individual windfarm asset.

        Parameters
        ----------
        env : WombatEnvironment
            The simulation environment.
        repair_manager : RepairManager
            The simulation repair and maintenance task manager.
        t_id : str
            The unique identifier for the asset.
        name : str
            The long form name/descriptor for the system/asset.
        subassemblies : dict
            The dictionary of subassemblies required for the system/asset.
        system : str
            The identifier should be one of "turbine" or "substation" to indicate the
            type of system this will be.

        Raises
        ------
        ValueError
            [description]
        """
        self.env = env
        self.repair_manager = repair_manager
        self.id = t_id
        self.name = name
        self.servicing = False
        self.cable_failure = False
        self.capacity = subassemblies["capacity_kw"]

        system = system.lower().strip()
        self._calculate_system_value(subassemblies)
        if system == "turbine":
            self._create_turbine_subassemblies(subassemblies)
            self._initialize_power_curve(subassemblies.get("power_curve", None))
        elif system == "substation":
            self._create_substation_subassemblies(subassemblies)
        else:
            raise ValueError("'system' must be one of 'turbine' or 'substation'!")

    def _calculate_system_value(self, subassemblies: dict) -> None:
        """Calculates the turbine's value based its capex_kw and capacity.

        Parameters
        ----------
        system : str
            One of "turbine" or "substation".
        subassemblies : dict
            Dictionary of subassemblies.
        """
        self.value = subassemblies["capacity_kw"] * subassemblies["capex_kw"]

    def _create_turbine_subassemblies(self, subassembly_data: dict) -> None:
        """Creates each subassembly as a separate attribute and also a list for quick
        access.

        Parameters
        ----------
        subassembly_data : dict
            Dictionary providing the maintenance and failure definitions for at least
            one of the following subassemblies:
             - electrical_system
             - electronic_control
             - sensors
             - hydraulic_system
             - yaw_system
             - rotor_blades
             - mechanical_brake
             - rotor_hub
             - gearbox
             - generator
             - supporting_structure
             - drive_train
        """
        subassembly_list = [
            "electrical_system",
            "electronic_control",
            "sensors",
            "hydraulic_system",
            "yaw_system",
            "rotor_blades",
            "mechanical_brake",
            "rotor_hub",
            "gearbox",
            "generator",
            "supporting_structure",
            "drive_train",
        ]

        # define the subassemblies  source: http://www.ecs.umass.edu/~arwade/wind_reliability.pdf
        self.subassemblies: list[Subassembly] = []

        # Check if the subassembly definition has been provided in the defintion
        # file, and if not, then skip it and move to the next subassembly
        try:
            self.electrical_system = Subassembly(
                self,
                self.env,
                "electrical_system",
                subassembly_data["electrical_system"],
            )
            self.subassemblies.append(self.electrical_system)
        except KeyError:
            pass

        try:
            self.electronic_control = Subassembly(
                self,
                self.env,
                "electronic_control",
                subassembly_data["electronic_control"],
            )
            self.subassemblies.append(self.electronic_control)
        except KeyError:
            pass

        try:
            self.sensors = Subassembly(
                self, self.env, "sensors", subassembly_data["sensors"]
            )
            self.subassemblies.append(self.sensors)
        except KeyError:
            pass

        try:
            self.hydraulic_system = Subassembly(
                self, self.env, "hydraulic_system", subassembly_data["hydraulic_system"]
            )
            self.subassemblies.append(self.hydraulic_system)
        except KeyError:
            pass

        try:
            self.yaw_system = Subassembly(
                self, self.env, "yaw_system", subassembly_data["yaw_system"]
            )
            self.subassemblies.append(self.yaw_system)
        except KeyError:
            pass

        try:
            self.rotor_blades = Subassembly(
                self, self.env, "rotor_blades", subassembly_data["rotor_blades"]
            )
            self.subassemblies.append(self.rotor_blades)
        except KeyError:
            pass

        try:
            self.mechanical_brake = Subassembly(
                self, self.env, "mechanical_brake", subassembly_data["mechanical_brake"]
            )
            self.subassemblies.append(self.mechanical_brake)
        except KeyError:
            pass

        try:
            self.rotor_hub = Subassembly(
                self, self.env, "rotor_hub", subassembly_data["rotor_hub"]
            )
            self.subassemblies.append(self.rotor_hub)
        except KeyError:
            pass

        try:
            self.gearbox = Subassembly(
                self, self.env, "gearbox", subassembly_data["gearbox"]
            )
            self.subassemblies.append(self.gearbox)
        except KeyError:
            pass

        try:
            self.generator = Subassembly(
                self, self.env, "generator", subassembly_data["generator"]
            )
            self.subassemblies.append(self.generator)
        except KeyError:
            pass

        try:
            self.supporting_structure = Subassembly(
                self,
                self.env,
                "supporting_structure",
                subassembly_data["supporting_structure"],
            )
            self.subassemblies.append(self.supporting_structure)
        except KeyError:
            pass

        try:
            self.drive_train = Subassembly(
                self, self.env, "drive_train", subassembly_data["drive_train"]
            )
            self.subassemblies.append(self.drive_train)
        except KeyError:
            pass

        if self.subassemblies == []:
            raise ValueError(
                "No subassembly data provided in the turbine configuration. At least"
                " defintion for the following subassemblies must be provided:"
                f" {subassembly_list}"
            )

        self.env.log_action(
            agent=self.name,
            action=f"subassemblies created: {[s.id for s in self.subassemblies]}",
            reason="windfarm initialization",
            system_id=self.id,
            system_name=self.name,
            system_ol=self.operating_level,
            part_ol=1,
            additional="initialization",
        )

    def _initialize_power_curve(self, power_curve_dict: dict) -> None:
        """Creates the power curve function based on the ``power_curve`` input in the
        ``subassembly_data`` dictionary. If there is no valid input, then 0 will always
        be reutrned.

        Parameters
        ----------
        power_curve_dict : dict
            The turbine definition dictionary.
        """
        self.power_curve: Callable
        if power_curve_dict is None:
            self.power_curve = IEC_power_curve(pd.Series([0]), pd.Series([0]))
        else:
            power_curve = power_curve_dict["file"]
            power_curve = self.env.data_dir / "windfarm" / power_curve
            power_curve = pd.read_csv(f"{power_curve}")
            power_curve = power_curve.loc[power_curve.power_kw != 0].reset_index(
                drop=True
            )
            bin_width = power_curve_dict.get("bin_width", 0.5)
            self.power_curve = IEC_power_curve(
                power_curve.windspeed_ms,
                power_curve.power_kw,
                windspeed_start=power_curve.windspeed_ms.min(),
                windspeed_end=power_curve.windspeed_ms.max(),
                bin_width=bin_width,
            )

    def _create_substation_subassemblies(self, subassembly_data: dict) -> None:
        """Creates each subassembly as a separate attribute and also a list for quick
        access.

        Parameters
        ----------
        subassembly_data : dict
            Dictionary containing the transformer maintenance and failure configurations.
        """
        self.subassemblies = []
        try:
            self.transformer = Subassembly(
                self, self.env, "transformer", subassembly_data["transformer"]
            )
            self.subassemblies = [self.transformer]
        except KeyError:
            pass

        if self.subassemblies == []:
            raise ValueError(
                "No subassembly data provided in the substation configuration; a"
                " transformer definition must be provided."
            )

        self.env.log_action(
            system_id=self.id,
            system_name=self.name,
            system_ol=self.operating_level,
            part_ol=1,
            agent=self.name,
            action="subassemblies created",
            reason="windfarm initialization",
            additional="initialization",
        )

    def interrupt_all_subassembly_processes(self) -> None:
        """Interrupts the running processes in all of the system's subassemblies."""
        [subassembly.interrupt_processes() for subassembly in self.subassemblies]  # type: ignore

    @property
    def operating_level(self) -> float:
        """The turbine's operating level, based on subassembly and cable performance.

        Returns
        -------
        float
            Operating level of the turbine.
        """
        if self.cable_failure or self.servicing:
            return 0.0
        else:
            return reduce(_product, [sub.operating_level for sub in self.subassemblies])

    @property
    def operating_level_wo_servicing(self) -> float:
        """The turbine's operating level, based on subassembly and cable performance,
        without accounting for servicing status.

        Returns
        -------
        float
            Operating level of the turbine.
        """
        if self.cable_failure:
            return 0.0
        else:
            return reduce(_product, [sub.operating_level for sub in self.subassemblies])

    def power(self, windspeed: list[float] | np.ndarray) -> np.ndarray:
        """Generates the power output for an iterable of windspeed values.

        Parameters
        ----------
        windspeed : list[float] | np.ndarrays
            Windspeed values, in m/s.

        Returns
        -------
        np.ndarray
            Power production, in kW.
        """
        return self.power_curve(windspeed)



"""
Initializes the simulation modules, classes, and functions.

isort:skip_file
"""

from .data_classes import (
    Maintenance,
    Failure,
    SubassemblyData,
    RepairRequest,
    ServiceEquipmentData,
    ScheduledServiceEquipmentData,
    UnscheduledServiceEquipmentData,
    FixedCosts,
    FromDictMixin,
)
from .environment import WombatEnvironment
from .post_processor import Metrics
from .repair_management import RepairManager
from .service_equipment import ServiceEquipment
from .simulation_api import Configuration, Simulation



"""Provides the Subassembly class"""


from typing import Generator  # type: ignore

import simpy  # type: ignore

from wombat.core import (
    Failure,
    Maintenance,
    RepairRequest,
    SubassemblyData,
    WombatEnvironment,
)


# TODO: Need a better method for checking if a repair has been made to bring the
# subassembly back online
HOURS = 8760
TIMEOUT = 24  # Wait time of 1 day for replacement to occur


class Subassembly:
    """A major system composes the turbine or substation objects."""

    def __init__(
        self,
        turbine,
        env: WombatEnvironment,
        s_id: str,
        subassembly_data: dict,
    ) -> None:
        """Creates a subassembly object that models various maintenance and failure types.

        Parameters
        ----------
        turbine : wombat.windfarm.System
            The turbine containing subassembly object(s).
        env : WombatEnvironment
            The simulation environment.
        s_id : str
            A unique identifier for the subassembly within the turbine.
        subassembly_data : dict
            A dictionary to be passed to ``SubassemblyData`` for creation and validation.
        """

        self.env = env
        self.turbine = turbine
        self.id = s_id

        subassembly_data = {**subassembly_data, "system_value": self.turbine.value}
        self.data = SubassemblyData.from_dict(subassembly_data)
        self.name = self.data.name

        self.broken = False
        self.operating_level = 1.0

        self.processes = dict(self._create_processes())

    def _create_processes(self):
        """Creates the processes for each of the failure and maintenance types.

        Yields
        -------
        Tuple[Union[str, int], simpy.events.Process]
            Creates a dictionary to keep track of the running processes within the
            subassembly.
        """
        for level, failure in self.data.failures.items():
            yield level, self.env.process(self.run_single_failure(failure))

        for i, maintenance in enumerate(self.data.maintenance):
            yield f"m{i}", self.env.process(self.run_single_maintenance(maintenance))

    def interrupt_processes(self) -> None:
        """Interrupts all of the running processes within the subassembly except for the
        process associated with failure that triggers the catastrophic failure.

        Parameters
        ----------
        subassembly : Subassembly
            The subassembly that should have all processes interrupted.
        """
        for _, process in self.processes.items():
            try:
                process.interrupt()
            except RuntimeError:
                # This error occurs for the process halting all other processes.
                pass

    def interrupt_all_subassembly_processes(self) -> None:
        """Thin wrapper for ``turbine.interrupt_all_subassembly_processes``."""
        self.turbine.interrupt_all_subassembly_processes()

    def run_single_maintenance(self, maintenance: Maintenance) -> Generator:
        """Runs a process to trigger one type of maintenance request throughout the simulation.

        Parameters
        ----------
        maintenance : Maintenance
            A maintenance category.

        Yields
        -------
        simpy.events.Timeout
            Time between maintenance requests.
        """
        while True:
            hours_to_next = maintenance.frequency
            if hours_to_next == 0:
                remainder = self.env.max_run_time - self.env.now
                try:
                    # TODO: determine if this logging is really needed
                    # self.env.log_action(
                    #     system_id=self.turbine.id,
                    #     system_name=self.turbine.name,
                    #     part_id=self.id,
                    #     part_name=self.name,
                    #     system_ol=self.turbine.operating_level,
                    #     part_ol=self.operating_level,
                    #     agent=self.name,
                    #     action="none",
                    #     reason=f"{self.name} is not modeled",
                    #     additional="no maintenance will be modeled for select subassembly",
                    #     duration=remainder,
                    # )
                    yield self.env.timeout(remainder)
                except simpy.Interrupt:
                    remainder -= self.env.now

            while hours_to_next > 0:
                try:
                    # If the replacement has not been completed, then wait another minute
                    if self.turbine.operating_level == 0 or self.turbine.servicing:
                        yield self.env.timeout(TIMEOUT)
                        continue

                    start = self.env.now
                    yield self.env.timeout(hours_to_next)
                    hours_to_next = 0

                    # Automatically submit a repair request
                    # NOTE: mypy is not caught up with attrs yet :(
                    repair_request = RepairRequest(  # type: ignore
                        system_id=self.turbine.id,
                        system_name=self.turbine.name,
                        subassembly_id=self.id,
                        subassembly_name=self.name,
                        severity_level=0,
                        details=maintenance,
                    )
                    repair_request = self.turbine.repair_manager.submit_request(
                        repair_request
                    )
                    self.env.log_action(
                        system_id=self.turbine.id,
                        system_name=self.turbine.name,
                        part_id=self.id,
                        part_name=self.name,
                        system_ol=self.turbine.operating_level,
                        part_ol=self.operating_level,
                        agent=self.name,
                        action="maintenance request",
                        reason=maintenance.description,
                        additional="request",
                        request_id=repair_request.request_id,
                    )

                except simpy.Interrupt:
                    if self.broken:
                        # The subassembly had so restart the maintenance cycle
                        hours_to_next = 0
                        continue
                    else:
                        # A different subassembly failed so we need to subtract the
                        # amount of passed time
                        hours_to_next -= self.env.now - start

    def run_single_failure(self, failure: Failure) -> Generator:
        """Runs a process to trigger one type of failure repair request throughout the simulation.

        Parameters
        ----------
        failure : Failure
            A failure classification.

        Yields
        -------
        simpy.events.Timeout
            Time between failure events that need to request a repair.
        """
        while True:
            hours_to_next = failure.hours_to_next_failure()
            if hours_to_next is None:
                remainder = self.env.max_run_time - self.env.now
                try:
                    # TODO: determine if this logging is really needed
                    # self.env.log_action(
                    #     system_id=self.turbine.id,
                    #     system_name=self.turbine.name,
                    #     part_id=self.id,
                    #     part_name=self.name,
                    #     system_ol=self.turbine.operating_level,
                    #     part_ol=self.operating_level,
                    #     agent=self.name,
                    #     action="none",
                    #     reason=f"{self.name} is not modeled",
                    #     additional="no failures will be modeled for select subassembly",
                    #     duration=remainder,
                    # )
                    yield self.env.timeout(remainder)
                except simpy.Interrupt:
                    remainder -= self.env.now

            else:
                while hours_to_next > 0:  # type: ignore
                    try:
                        if self.turbine.operating_level == 0 or self.turbine.servicing:
                            yield self.env.timeout(TIMEOUT)
                            continue

                        start = self.env.now
                        yield self.env.timeout(hours_to_next)
                        hours_to_next = 0
                        self.operating_level *= 1 - failure.operation_reduction
                        if failure.operation_reduction == 1:
                            self.broken = True
                            self.interrupt_all_subassembly_processes()

                            # Remove previously submitted requests as a replacement is required
                            _ = self.turbine.repair_manager.purge_subassembly_requests(
                                self.turbine.id, self.id
                            )

                        # Automatically submit a repair request
                        # NOTE: mypy is not caught up with attrs yet :(
                        repair_request = RepairRequest(  # type: ignore
                            self.turbine.id,
                            self.turbine.name,
                            self.id,
                            self.name,
                            failure.level,
                            failure,
                        )
                        repair_request = self.turbine.repair_manager.submit_request(
                            repair_request
                        )

                        self.env.log_action(
                            system_id=self.turbine.id,
                            system_name=self.turbine.name,
                            part_id=self.id,
                            part_name=self.name,
                            system_ol=self.turbine.operating_level,
                            part_ol=self.operating_level,
                            agent=self.name,
                            action="repair request",
                            reason=failure.description,
                            additional=f"severity level {failure.level}",
                            request_id=repair_request.request_id,
                        )

                    except simpy.Interrupt:
                        if self.broken:
                            # The subassembly had to be replaced so the timing to next failure
                            # will reset
                            hours_to_next = 0
                            continue
                        else:
                            # A different subassembly failed so we need to subtract the
                            # amount of passed time
                            hours_to_next -= self.env.now - start
